{% extends "base.html" %}

{% block content %}
    <h5 class="card-header">Gilda</h5>
    <div class="card-body">
        <p>
            This is the Gilda grounding service.
        </p>
        {{ wtf.quick_form(form, form_type='horizontal') }}
    </div>
{% endblock %}
