# -*- coding: utf-8 -*-

"""High-level API for accessing content."""

from .alts import (  # noqa: F401
    get_alts_to_id,
    get_id_to_alts,
    get_primary_curie,
    get_primary_identifier,
)
from .hierarchy import (  # noqa: F401
    get_ancestors,
    get_descendants,
    get_hierarchy,
    get_subhierarchy,
    has_ancestor,
    is_descendent,
)
from .metadata import get_metadata  # noqa: F401
from .names import (  # noqa: F401
    get_definition,
    get_id_definition_mapping,
    get_id_name_mapping,
    get_id_synonyms_mapping,
    get_ids,
    get_name,
    get_name_by_curie,
    get_name_id_mapping,
    get_synonyms,
)
from .properties import (  # noqa: F401
    get_filtered_properties_df,
    get_filtered_properties_mapping,
    get_filtered_properties_multimapping,
    get_properties,
    get_properties_df,
    get_property,
)
from .relations import (  # noqa: F401
    get_filtered_relations_df,
    get_id_multirelations_mapping,
    get_relation,
    get_relation_mapping,
    get_relations_df,
)
from .species import get_id_species_mapping, get_species  # noqa: F401
from .typedefs import get_typedef_df  # noqa: F401
from .xrefs import (  # noqa: F401
    get_filtered_xrefs,
    get_sssom_df,
    get_xref,
    get_xrefs,
    get_xrefs_df,
)
